
import os
import shutil

from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint
from pytorch_lightning.logging import TestTubeLogger
from torchline.utils import Logger

from .build import TRAINER_REGISTRY

__all__ = [
    'DefaultTrainer'
]


@TRAINER_REGISTRY.register()
class DefaultTrainer(Trainer):
    def __init__(self, cfg, hparams):
        self.cfg = cfg
        self.hparams = hparams

        self.logger = self._logger()
        self.logger_print = Logger(__name__, cfg).getlogger()
        self.early_stop_callback = self._early_stop_callback()
        self.checkpoint_callback = self._checkpoint_callback()

        # you can update trainer_params to change different parameters
        self.trainer_params = dict(self.cfg.trainer)
        self.trainer_params.update({
            'logger': self.logger,
            'early_stop_callback': self.early_stop_callback,
            'checkpoint_callback': self.checkpoint_callback,
        })
        self.trainer_params.pop('name')
        for key in self.trainer_params:
            self.trainer_params[key] = self.parse_cfg_param(self.trainer_params[key])
            
        super(DefaultTrainer, self).__init__(**self.trainer_params)

    def parse_cfg_param(self, cfg_item):
        return cfg_item if cfg_item not in ['', []] else None

    def _logger(self):

        # logger
        logger_cfg = self.cfg.trainer.logger
        assert logger_cfg.setting in [0,1,2], "You can only set three logger levels [0,1,2], but you set {}".format(logger_cfg.setting)
        if logger_cfg.type == 'mlflow':
            raise NotImplementedError
            # params = {key: logger_cfg.mlflow[key] for key in logger_cfg.mlflow}
            # custom_logger = MLFlowLogger(**params)
        elif logger_cfg.type == 'test_tube':
            params = {key: logger_cfg.test_tube[key] for key in logger_cfg.test_tube} # key: save_dir, name, version

            # save_dir: logger root path: 
            if self.cfg.trainer.default_save_path:
                save_dir = self.cfg.trainer.default_save_path
            else:
                save_dir = logger_cfg.test_tube.save_dir
            
            # version
            if logger_cfg.setting==1:
                return False
            else:
                version = self.cfg.log.path[-1]

            default_logger = TestTubeLogger(save_dir, name='torchline_logs', version=version)
            params.update({'version': version, 'name':logger_cfg.test_tube.name, 'save_dir': save_dir})
            custom_logger = TestTubeLogger(**params)
        else:
            print(f"{logger_cfg.type} not supported")
            raise NotImplementedError
        
        loggers = {
            0: default_logger,
            1: False,
            2: custom_logger
        } # 0: True (default)  1: False  2: custom
        logger = loggers[logger_cfg.setting]

        return logger

    def _early_stop_callback(self):
        # early_stop_callback hooks
        hooks = self.cfg.hooks
        params = {key: hooks.early_stopping[key] for key in hooks.early_stopping if key != 'setting'}
        early_stop_callbacks = {
            0: True,  # default setting
            1: False, # do not use early stopping
            2: EarlyStopping(**params) # use custom setting
        }
        assert hooks.early_stopping.setting in early_stop_callbacks, 'The setting of early stopping can only be in [0,1,2]'
        early_stop_callback = early_stop_callbacks[hooks.early_stopping.setting]
        return early_stop_callback

    def _checkpoint_callback(self):
        # checkpoint_callback hooks
        hooks = self.cfg.hooks
        assert hooks.model_checkpoint.setting in [0,1,2], "You can only set three ckpt levels [0,1,2], but you set {}".format(hooks.model_checkpoint.setting)

        params = {key: hooks.model_checkpoint[key] for key in hooks.model_checkpoint if key != 'setting'}
        if hooks.model_checkpoint.setting==2:
            if hooks.model_checkpoint.filepath.strip()=='':
                filepath = os.path.join(self.cfg.log.path,'checkpoints')
                params.update({'filepath': filepath})
            else:
                self.logger_print.warn("The specified checkpoint path is not recommended!")
        checkpoint_callbacks = {
            0: True,
            1: False,
            2: ModelCheckpoint(**params)
        }
        checkpoint_callback = checkpoint_callbacks[hooks.model_checkpoint.setting]
        return checkpoint_callback



#coding=utf-8
import torch
import torch.nn.functional as F
from .build import LOSS_FN_REGISTRY

__all__ = [
    'CrossEntropy',
    'CrossEntropyLabelSmooth'
]

@LOSS_FN_REGISTRY.register()
def CrossEntropy(cfg):
    weight = cfg.loss.class_weight
    if weight in ['', None, []]:
        weight = None
    else:
        weight = torch.tensor(weight)
        if torch.cuda.is_available(): weight=weight.cuda()
    return torch.nn.CrossEntropyLoss(weight=weight)

@LOSS_FN_REGISTRY.register()
class CrossEntropyLabelSmooth(torch.nn.Module):
    def __init__(self, cfg):
        super(CrossEntropyLabelSmooth, self).__init__()
        try:
            self.label_smoothing = cfg.loss.label_smoothing
        except:
            self.label_smoothing = 0.1

    def forward(self, pred, target):
        logsoftmax = torch.nn.LogSoftmax()
        n_classes = pred.size(1)
        # convert to one-hot
        target = torch.unsqueeze(target, 1)
        soft_target = torch.zeros_like(pred)
        soft_target.scatter_(1, target, 1)
        # label smoothing
        soft_target = soft_target * (1 - self.label_smoothing) + self.label_smoothing / n_classes
        return torch.mean(torch.sum(- soft_target * logsoftmax(pred), 1))


from albumentations import (CLAHE, Blur, ChannelDropout, ChannelShuffle,
                            Compose, Cutout, Flip, GaussNoise, GridDistortion,
                            HueSaturationValue, IAAAdditiveGaussianNoise,
                            IAAEmboss, MotionBlur, Normalize, OneOf,
                            OpticalDistortion, RandomBrightnessContrast,
                            RandomGridShuffle, Resize, ShiftScaleRotate)
from albumentations.pytorch.transforms import ToTensor

from .transforms import TRANSFORMS_REGISTRY, BaseTransforms

__all__ = [
    'AlbumentationsTransforms'
]

@TRANSFORMS_REGISTRY.register()
class AlbumentationsTransforms(BaseTransforms):
    def __init__(self, cfg):
        super(AlbumentationsTransforms, self).__init__(cfg)
        self.is_train = cfg.dataset.is_train
        self.img_size = cfg.input.size
        self.mean = cfg.transforms.tensor.normalization.mean
        self.std = cfg.transforms.tensor.normalization.std
        self.normalize = Normalize(self.mean, self.std)
        self.init_transforms = self.get_transform()
        self.transform = self.parse_transform
    
    def parse_transform(self, image):
        image = self.init_transforms(image=image)['image']
        return image

    @property
    def valid_transform(self):
        height, width = self.img_size
        return Compose([Resize(height, width),
            self.normalize,
            ToTensor()])

    @property
    def train_transform(self):
        # albumentations transforms cfg
        at_cfg = self.cfg.abtfs
        height, width = self.img_size
        transforms_list = [Resize(height, width), Flip()]
        
        # random_grid_shuffle
        if at_cfg.random_grid_shuffle.enable:
            grid = at_cfg.random_grid_shuffle.grid
            grid = (grid,grid)
            transforms_list.append(RandomGridShuffle((grid)))

        # channel_shuffle
        if at_cfg.channel_shuffle.enable:
            transforms_list.append(ChannelShuffle(p=1))
        
        # channel_dropout
        if at_cfg.channel_dropout.enable:
            drop_range = at_cfg.channel_dropout.drop_range
            fill_value = at_cfg.channel_dropout.fill_value
            transforms_list.append(ChannelDropout(drop_range, fill_value, p=1))

        # noise
        if at_cfg.noise.enable:
            transforms_list.append(OneOf([
                IAAAdditiveGaussianNoise(),
                GaussNoise(),
            ], p=1))

        # blur
        if at_cfg.blur.enable:
            transforms_list.append(OneOf([
                MotionBlur(),
                Blur(blur_limit=3,),
            ], p=1))

        # rotate
        if at_cfg.rotate.enable:
            transforms_list.append(ShiftScaleRotate(shift_limit=0.0625, scale_limit=0.2,
                                                    rotate_limit=45, p=1))

        # distortion
        if at_cfg.distortion.enable:
            transforms_list.append(OneOf([
                OpticalDistortion(p=0.3),
                GridDistortion(p=.3),
            ], p=1))

        # bright
        if at_cfg.bright.enable:
            transforms_list.append(
                OneOf([
                    CLAHE(clip_limit=2),
                    RandomBrightnessContrast(p=0.8),            
                ], p=1))

        # hue color
        if at_cfg.hue.enable:
            transforms_list.append(HueSaturationValue(p=0.3))
        
        # cutout
        if at_cfg.cutout.enable:
            num_holes = at_cfg.cutout.num_holes
            size = at_cfg.cutout.size
            fill_value = at_cfg.cutout.fill_value
            transforms_list.append(Cutout(num_holes, size, size, fill_value, 1))
        transforms_list.append(self.normalize)
        transforms_list.append(ToTensor())

        return Compose(transforms_list)



import os
import shutil

import torch
from yacs.config import CfgNode as _CfgNode

from torchline.utils import Logger


class CfgNode(_CfgNode):

    def _version(self):
        '''
        calculate the version of the configuration and logger
        '''
        def calc_version(path):
            if (not os.path.exists(path)) or len(os.listdir(path))==0:
                version = 0
            else:
                versions = [int(v.split('_')[-1]) for v in os.listdir(path)]
                version = max(versions)+1
            return version

        save_dir=self.trainer.default_save_path
        logger_name=self.trainer.logger.test_tube.name
        path = os.path.join(save_dir, logger_name)
        if self.trainer.logger.setting==0:
            version = calc_version(path)
        elif self.trainer.logger.setting==2:
            if self.trainer.logger.test_tube.version<0:
                version = calc_version(path)
            else:
                version = int(self.trainer.logger.test_tube.version)
        return version
    
    def setup_cfg_with_hparams(self, hparams):
        """
        Create configs and perform basic setups. 
        Args:
            hparams: arguments args from command line
        """
        self.merge_from_file(hparams.config_file)
        self.merge_from_list(hparams.opts)
        self.update({'hparams': hparams})

        # './outputs/torchline_logs/version_0/checkpoints/_ckpt_epoch_1.ckpt'
        ckpt_file = self.trainer.resume_from_checkpoint
        if ckpt_file:
            assert os.path.exists(ckpt_file), f"{ckpt_file} not exits"
            ckpt_path = os.path.dirname(ckpt_file).split('/')[:-1]
            self.log.path = ''.join([p+'/' for p in ckpt_path])
            self.log.name = os.path.join(self.log.path, 'log.txt')
        else:
            version = self._version()
            save_dir = self.trainer.default_save_path
            logger_name = self.trainer.logger.test_tube.name
            self.log.path = os.path.join(save_dir, logger_name, f"version_{version}")
            self.log.name = os.path.join(self.log.path, 'log.txt')
        self.freeze()

        os.makedirs(self.log.path, exist_ok=True)
        
        # log.txt
        logger_print = Logger(__name__, self).getlogger()

        # copy config file
        src_cfg_file = hparams.config_file # source config file
        cfg_file_name = os.path.basename(src_cfg_file) # config file name
        dst_cfg_file = os.path.join(self.log.path, cfg_file_name)
        with open(dst_cfg_file, 'w') as f:
            hparams = self.hparams
            self.pop('hparams')
            f.write(str(self))
            self.update({'hparams': hparams})

        if not (hparams.test_only or hparams.predict_only):
            torch.backends.cudnn.benchmark = False
        else:
            torch.backends.cudnn.benchmark = self.DEFAULT_CUDNN_BENCHMARK

        logger_print.info("Running with full config:\n{}".format(self))

    def __str__(self):
        def _indent(s_, num_spaces):
            s = s_.split("\n")
            if len(s) == 1:
                return s_
            first = s.pop(0)
            s = [(num_spaces * " ") + line for line in s]
            s = "\n".join(s)
            s = first + "\n" + s
            return s

        r = ""
        s = []
        for k, v in sorted(self.items()):
            seperator = "\n" if isinstance(v, CfgNode) else " "
            v = f"'{v}'" if isinstance(v, str) else v
            attr_str = "{}:{}{}".format(str(k), seperator, str(v))
            attr_str = _indent(attr_str, 4)
            s.append(attr_str)
        r += "\n".join(s)
        return r

global_cfg = CfgNode()

def get_cfg():
    '''
    Get a copy of the default config.

    Returns:
        a CfgNode instance.
    '''
    from .default import _C
    return _C.clone()


'''ResNet in PyTorch.

For Pre-activation ResNet, see 'preact_resnet.py'.

Reference:
[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
    Deep Residual Learning for Image Recognition. arXiv:1512.03385
'''
import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import models
from torchvision.models.utils import load_state_dict_from_url
from torchvision.models.resnet import BasicBlock, Bottleneck

from .build import META_ARCH_REGISTRY

__all__ = [
    'ResNet',
    'Resnet18',
    'Resnet34',
    'Resnet50',
    'Resnet101',
    'Resnet152',
    'Resnext50_32x4d',
    'Resnext101_32x8d',
    'Wide_resnet50_2',
    'Wide_resnet101_2'
]

model_urls = {
    'resnet18': 'https://download.pytorch.org/models/resnet18-5c106cde.pth',
    'resnet34': 'https://download.pytorch.org/models/resnet34-333f7ec4.pth',
    'resnet50': 'https://download.pytorch.org/models/resnet50-19c8e357.pth',
    'resnet101': 'https://download.pytorch.org/models/resnet101-5d3b4d8f.pth',
    'resnet152': 'https://download.pytorch.org/models/resnet152-b121ed2d.pth',
    'resnext50_32x4d': 'https://download.pytorch.org/models/resnext50_32x4d-7cdf4587.pth',
    'resnext101_32x8d': 'https://download.pytorch.org/models/resnext101_32x8d-8ba56ff5.pth',
    'wide_resnet50_2': 'https://download.pytorch.org/models/wide_resnet50_2-95faca4d.pth',
    'wide_resnet101_2': 'https://download.pytorch.org/models/wide_resnet101_2-32ee1156.pth',
}

class ResNet(models.ResNet):
    def __init__(self, block, num_blocks, classes=10, **kwargs):
        super(ResNet, self).__init__(block, num_blocks, **kwargs)

    def load_pretrained_weights(self, name, expansion, classes, pretrained=False):
        '''pretrained
        Args:
            name: model name, e.g. 'resnet18'
            expansion: block expansion
            classes: the number of classes
        '''
        if pretrained:
            state_dict = load_state_dict_from_url(model_urls[name], progress=True)
            self.load_state_dict(state_dict)
        if classes != 1000:
            self.fc = nn.Linear(512*expansion, classes, bias=False)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)
        return x

@META_ARCH_REGISTRY.register()
class Resnet18(ResNet):
    def __init__(self, cfg):
        classes = cfg.model.classes
        super(Resnet18, self).__init__(BasicBlock, [2,2,2,2], classes=classes)
        self.load_pretrained_weights('resnet18', BasicBlock.expansion, classes, cfg.model.pretrained)
        
@META_ARCH_REGISTRY.register()
class Resnet34(ResNet):
    def __init__(self, cfg):
        classes = cfg.model.classes
        super(Resnet34, self).__init__(BasicBlock, [3,4,6,3], classes=classes)
        self.load_pretrained_weights('resnet50', BasicBlock.expansion, classes, cfg.model.pretrained)

@META_ARCH_REGISTRY.register()
class Resnet50(ResNet):
    def __init__(self, cfg):
        classes = cfg.model.classes
        super(Resnet50, self).__init__(Bottleneck, [3,4,6,3], classes=classes)
        self.load_pretrained_weights('resnet50', Bottleneck.expansion, classes, cfg.model.pretrained)

@META_ARCH_REGISTRY.register()
class Resnet101(ResNet):
    def __init__(self, cfg):
        classes = cfg.model.classes
        super(Resnet101, self).__init__(Bottleneck, [3,4,23,3], classes=classes)
        self.load_pretrained_weights('resnet101', Bottleneck.expansion, classes, cfg.model.pretrained)

@META_ARCH_REGISTRY.register()
class Resnet152(ResNet):
    def __init__(self, cfg):
        classes = cfg.model.classes
        super(Resnet152, self).__init__(Bottleneck, [3,8,36,3], classes=classes)
        self.load_pretrained_weights('resnet152', Bottleneck.expansion, classes, cfg.model.pretrained)


@META_ARCH_REGISTRY.register()
class Resnext50_32x4d(ResNet):
    def __init__(self, cfg):
        classes = cfg.model.classes
        super(Resnext50_32x4d, self).__init__(Bottleneck, [3, 4, 6, 3], classes=classes, groups=32, width_per_group=4)
        self.load_pretrained_weights('resnext50_32x4d', Bottleneck.expansion, classes, cfg.model.pretrained)

@META_ARCH_REGISTRY.register()
class Resnext101_32x8d(ResNet):
    def __init__(self, cfg):
        classes = cfg.model.classes
        super(Resnext101_32x8d, self).__init__(Bottleneck, [3, 4, 23, 3], classes=classes, groups=32, width_per_group=8)
        self.load_pretrained_weights('resnext101_32x8d', Bottleneck.expansion, classes, cfg.model.pretrained)

@META_ARCH_REGISTRY.register()
class Wide_resnet50_2(ResNet):
    def __init__(self, cfg):
        classes = cfg.model.classes
        super(Wide_resnet50_2, self).__init__(Bottleneck, [3, 4, 6, 3], classes=classes, width_per_group=64 * 2)
        self.load_pretrained_weights('wide_resnet50_2', Bottleneck.expansion, classes, cfg.model.pretrained)

@META_ARCH_REGISTRY.register()
class Wide_resnet101_2(ResNet):
    def __init__(self, cfg):
        classes = cfg.model.classes
        super(Wide_resnet101_2, self).__init__(Bottleneck, [3, 4, 23, 3], classes=classes, width_per_group=64 * 2)
        self.load_pretrained_weights('wide_resnet101_2', Bottleneck.expansion, classes, cfg.model.pretrained)


# demo: customize the forward process
# class Net(Resnet18):
#     def __init__(self, cfg):
#         super(Net, self).__init__(cfg)

#     def forward(self, x):
#         x = self.conv1(x)
#         x = self.bn1(x)
#         x = self.relu(x)
#         x = self.maxpool(x)

#         x = self.layer1(x)
#         x = self.layer2(x)
#         x = self.layer3(x)
#         x = self.layer4(x)
#         print("Hello")
#         x = self.avgpool(x)
#         x = torch.flatten(x, 1)
#         x = self.fc(x)
#         return x

def test():
    import sys
    sys.path.append('..')
    from torchline.config import get_cfg
    from torchline.models import build_model
    cfg = get_cfg()
    cfg.model.name = 'Resnet18'
    cfg.model.pretrained = False
    cfg.model.classes = 10
    print(cfg.model)
    net = build_model(cfg)
    y = net(torch.randn(1,3,64,64))
    print(y.size())

if __name__ == '__main__':
    test()



import types

import torch
import torch.nn as nn
from torchvision import models
from torchvision.models.mobilenet import MobileNetV2

from .build import META_ARCH_REGISTRY

__all__ = [
    'MobileNetV2',
    'MobileNet_V2'
]


def modify_mobilenet(model):
    # Modify attributs
    model.last_linear = model.classifier
    model.classifier = None

    def logits(self, features):
        x = features.mean([2, 3])
        x = self.last_linear(x)
        return x

    def forward(self, input):
        x = self.features(input)
        x = self.logits(x)
        return x

    # Modify methods
    model.logits = types.MethodType(logits, model)
    model.forward = types.MethodType(forward, model)
    return model

def generate_mobilenet(cfg, name):
    pretrained=cfg.model.pretrained
    classes = cfg.model.classes
    if 'dropout' in cfg.model:
        dropout = cfg.model.dropout
    else:
        dropout = 0.2
    model = eval(f"models.{name}(pretrained={pretrained})")
    if classes != 1000:
        in_features = model.classifier[1].in_features
        model.classifier = nn.Sequential(
            nn.Dropout(p=dropout, inplace=True),
            nn.Linear(in_features, classes, bias=False))
    return modify_mobilenet(model)

@META_ARCH_REGISTRY.register()
def MobileNet_V2(cfg):
    return generate_mobilenet(cfg, 'mobilenet_v2')


#!/usr/bin/python
# -*- coding:utf-8 -*-

import logging
import time
import os

__all__ = [
    'Logger'
]

class Logger(object):
    def __init__(self, logger_name=None, cfg=None):
        '''
            指定保存日志的文件路径，日志级别，以及调用文件
            将日志存入到指定的文件中
        '''

        # 创建一个logger
        if cfg is None:
            file = 'log.txt'
        else:
            file = cfg.log.name
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(logging.INFO)
        self.logger.propagate = False
        if (self.logger.hasHandlers()):
            self.logger.handlers.clear()
        formatter = logging.Formatter('[%(asctime)s] %(filename)s->%(funcName)s line:%(lineno)d [%(levelname)s]%(message)s')

        if file:
            hdlr = logging.FileHandler(file, 'a', encoding='utf-8')
            hdlr.setLevel(logging.INFO)
            hdlr.setFormatter(formatter)
            self.logger.addHandler(hdlr)

        strhdlr = logging.StreamHandler()
        strhdlr.setLevel(logging.INFO)
        strhdlr.setFormatter(formatter)
        self.logger.addHandler(strhdlr)

        if file: hdlr.close()
        strhdlr.close()

    def getlogger(self):
        return self.logger



import types

import torch
import torch.nn as nn
from torchvision import models
from torchvision.models.resnet import ResNet

from .build import META_ARCH_REGISTRY

__all__ = [
    'ResNet',
    'Resnet18',
    'Resnet34',
    'Resnet50',
    'Resnet101',
    'Resnet152',
    'Resnext50_32x4d',
    'Resnext101_32x8d',
    'Wide_resnet50_2',
    'Wide_resnet101_2'
]


def modify_resnets(model):
    # Modify attributs
    model.last_linear = model.fc
    model.fc = None

    def features(self, input):
        x = self.conv1(input)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        return x

    def logits(self, features):
        x = self.avgpool(features)
        x = x.view(x.size(0), -1)
        x = self.last_linear(x)
        return x

    def forward(self, input):
        x = self.features(input)
        x = self.logits(x)
        return x

    # Modify methods
    model.features = types.MethodType(features, model)
    model.logits = types.MethodType(logits, model)
    model.forward = types.MethodType(forward, model)
    return model

def generate_resnet(cfg, name):
    pretrained=cfg.model.pretrained
    classes = cfg.model.classes
    model = eval(f"models.{name}(pretrained={pretrained})")
    if classes != 1000:
        in_features = model.fc.in_features
        model.fc = nn.Linear(in_features, classes, bias=False)
    return modify_resnets(model)

@META_ARCH_REGISTRY.register()
def Resnet18(cfg):
    return generate_resnet(cfg, 'resnet18')

@META_ARCH_REGISTRY.register()
def Resnet34(cfg):
    return generate_resnet(cfg, 'resnet34')

@META_ARCH_REGISTRY.register()
def Resnet50(cfg):
    return generate_resnet(cfg, 'resnet50')

@META_ARCH_REGISTRY.register()
def Resnet101(cfg):
    return generate_resnet(cfg, 'resnet101')

@META_ARCH_REGISTRY.register()
def Resnet152(cfg):
    return generate_resnet(cfg, 'resnet152')

@META_ARCH_REGISTRY.register()
def Resnext50_32x4d(cfg):
    return generate_resnet(cfg, 'resnext50_32x4d')

@META_ARCH_REGISTRY.register()
def Resnext101_32x8d(cfg):
    return generate_resnet(cfg, 'resnext101_32x8d')

@META_ARCH_REGISTRY.register()
def Wide_resnet50_2(cfg):
    return generate_resnet(cfg, 'wide_resnet50_2')

@META_ARCH_REGISTRY.register()
def Wide_resnet101_2(cfg):
    return generate_resnet(cfg, 'wide_resnet101_2')


# demo: customize the forward process
def customize():
    import type
    res = models.resnet18(pretrained=False)
    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)
        print("Hello")
        return x
    res.forward = forward
    x = torch.rand(1,3,64,64)
    y = res(res, x)
    print(y.shape)


from torchline.utils import Registry

DATASET_REGISTRY = Registry("DATASET")
DATASET_REGISTRY.__doc__ = """
Registry for dataset, i.e. torch.utils.data.Dataset.

The registered object will be called with `obj(cfg)`
"""

__all__ = [
    'DATASET_REGISTRY',
    'build_data'
]

def build_data(cfg):
    """
    Built the dataset, defined by `cfg.dataset.name`.
    """
    name = cfg.dataset.name
    return DATASET_REGISTRY.get(name)(cfg)



'''EfficientNet in PyTorch.

Paper: "EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks".
'''
import torch
import torch.nn as nn
import torch.nn.functional as F


from .build import META_ARCH_REGISTRY

__all__ = [
    'EfficientNet',
    'EfficientNetB0',
]

class Block(nn.Module):
    '''expand + depthwise + pointwise + squeeze-excitation'''

    def __init__(self, in_planes, out_planes, expansion, stride):
        super(Block, self).__init__()
        self.stride = stride

        planes = expansion * in_planes
        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3,
                               stride=stride, padding=1, groups=planes, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(
            planes, out_planes, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn3 = nn.BatchNorm2d(out_planes)

        self.shortcut = nn.Sequential()
        if stride == 1 and in_planes != out_planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, out_planes, kernel_size=1,
                          stride=1, padding=0, bias=False),
                nn.BatchNorm2d(out_planes),
            )

        # SE layers
        self.fc1 = nn.Conv2d(out_planes, out_planes//16, kernel_size=1)
        self.fc2 = nn.Conv2d(out_planes//16, out_planes, kernel_size=1)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        shortcut = self.shortcut(x) if self.stride == 1 else out
        # Squeeze-Excitation
        w = F.avg_pool2d(out, out.size(2))
        w = F.relu(self.fc1(w))
        w = self.fc2(w).sigmoid()
        out = out * w + shortcut
        return out

@META_ARCH_REGISTRY.register()
class EfficientNet(nn.Module):
    def __init__(self, layers, num_classes=10):
        super(EfficientNet, self).__init__()
        self.layers = layers
        self.conv1 = nn.Conv2d(3, 32, kernel_size=3,
                               stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(32)
        self.layers = self._make_layers(in_planes=32)
        self.last_linear = nn.Linear(layers[-1][1], num_classes)

    def _make_layers(self, in_planes):
        layers = []
        for expansion, out_planes, num_blocks, stride in self.layers:
            strides = [stride] + [1]*(num_blocks-1)
            for stride in strides:
                layers.append(Block(in_planes, out_planes, expansion, stride))
                in_planes = out_planes
        return nn.Sequential(*layers)

    def features(self, input): # 2*3*32*32
        x = F.relu(self.bn1(self.conv1(input))) # 2*32*32*32
        x = self.layers(x) # 2*320*1*1
        return x

    def logits(self, features):
        x = nn.AdaptiveAvgPool2d((1,1))(features) # 2*320*1*1
        x = x.view(x.size(0), -1)
        x = self.last_linear(x)
        return x

    def forward(self, input): # 2*3*32*32
        x = self.features(input) # 2*320*1*1
        x = self.logits(x) # 2*num_classes
        return x


@META_ARCH_REGISTRY.register()
class EfficientNetB0(EfficientNet):
    def __init__(self, cfg):
        num_classes = cfg.model.classes
        # (expansion, out_planes, num_blocks, stride)
        layers = [(1,  16, 1, 2),
            (6,  24, 2, 1),
            (6,  40, 2, 2),
            (6,  80, 3, 2),
            (6, 112, 3, 1),
            (6, 192, 4, 2),
            (6, 320, 1, 2)]
        super(EfficientNetB0, self).__init__(layers, num_classes)


import types

import torch
import torch.nn as nn
from torchvision import models
from torchvision.models.mnasnet import MNASNet

from .build import META_ARCH_REGISTRY

__all__ = [
    'MNASNet',
    'MNASNet0_5',
    'MNASNet0_75',
    'MNASNet1_0',
    'MNASNet1_3',
]

# layers = [
#     # First layer: regular conv.
#     nn.Conv2d(3, depths[0], 3, padding=1, stride=2, bias=False),
#     nn.BatchNorm2d(depths[0], momentum=_BN_MOMENTUM),
#     nn.ReLU(inplace=True),
#     # Depthwise separable, no skip.
#     nn.Conv2d(depths[0], depths[0], 3, padding=1, stride=1,
#                 groups=depths[0], bias=False),
#     nn.BatchNorm2d(depths[0], momentum=_BN_MOMENTUM),
#     nn.ReLU(inplace=True),
#     nn.Conv2d(depths[0], depths[1], 1, padding=0, stride=1, bias=False),
#     nn.BatchNorm2d(depths[1], momentum=_BN_MOMENTUM),
#     # MNASNet blocks: stacks of inverted residuals.
#     _stack(depths[1], depths[2], 3, 2, 3, 3, _BN_MOMENTUM), #layer1
#     _stack(depths[2], depths[3], 5, 2, 3, 3, _BN_MOMENTUM), #layer2
#     _stack(depths[3], depths[4], 5, 2, 6, 3, _BN_MOMENTUM), #layer3
#     _stack(depths[4], depths[5], 3, 1, 6, 2, _BN_MOMENTUM), #layer4
#     _stack(depths[5], depths[6], 5, 2, 6, 4, _BN_MOMENTUM), #layer5
#     _stack(depths[6], depths[7], 3, 1, 6, 1, _BN_MOMENTUM), #layer6
#     # Final mapping to classifier input.
#     nn.Conv2d(depths[7], 1280, 1, padding=0, stride=1, bias=False),
#     nn.BatchNorm2d(1280, momentum=_BN_MOMENTUM),
#     nn.ReLU(inplace=True),
# ]
def modify_mnasnet(model):
    # Modify attributs
    model.stem = model.layers[:8]
    model.layer1 = model.layers[8]
    model.layer2 = model.layers[9]
    model.layer3 = model.layers[10]
    model.layer4 = model.layers[11]
    model.layer5 = model.layers[12]
    model.layer6 = model.layers[13]
    model.layer7 = model.layers[14:]
    model.last_linear = model.classifier
    model.layers = None
    model.classifier = None

    def features(self, input): # b*3*64*64
        x = self.stem(input)   # b*16*32*32
        x = self.layer1(x)     # b*16*16*16
        x = self.layer2(x)     # b*24*8*8
        x = self.layer3(x)     # b*40*4*4
        x = self.layer4(x)     # b*48*4*4
        x = self.layer5(x)     # b*96*2*2
        x = self.layer6(x)     # b*160*2*2
        x = self.layer7(x)     # b*1280*2*2
        return x

    def logits(self, features):
        x = features.mean([2, 3])
        x = self.last_linear(x)
        return x

    def forward(self, input):
        x = self.features(input)
        x = self.logits(x)
        return x

    # Modify methods
    model.features = types.MethodType(features, model)
    model.logits = types.MethodType(logits, model)
    model.forward = types.MethodType(forward, model)
    return model

def generate_mnasnet(cfg, name):
    pretrained=cfg.model.pretrained
    classes = cfg.model.classes
    if 'dropout' in cfg.model:
        dropout = cfg.model.dropout
    else:
        dropout = 0.2
    model = eval(f"models.{name}(pretrained={pretrained})")
    if classes != 1000:
        in_features = model.classifier[1].in_features
        model.classifier = nn.Sequential(
            nn.Dropout(p=dropout, inplace=True),
            nn.Linear(in_features, classes, bias=False))
    return modify_mnasnet(model)

@META_ARCH_REGISTRY.register()
def MNASNet0_5(cfg):
    return generate_mnasnet(cfg, 'mnasnet0_5')
    
@META_ARCH_REGISTRY.register()
def MNASNet0_75(cfg):
    return generate_mnasnet(cfg, 'mnasnet0_75')
    
@META_ARCH_REGISTRY.register()
def MNASNet1_0(cfg):
    return generate_mnasnet(cfg, 'mnasnet1_0')
    
@META_ARCH_REGISTRY.register()
def MNASNet1_3(cfg):
    return generate_mnasnet(cfg, 'mnasnet1_3')


from setuptools import setup, find_packages

with open('requirements.txt', 'r') as f:
    requirements = f.read().splitlines()

setup(
    name="torchline", # Replace with your own username
    version="0.2.4.0",
    author="marsggbo",
    author_email="csxinhe@comp.hkbu.edu.hk",
    description="A framework for easy to use Pytorch",
    long_description='''
    The ML developer can easily use this framework to implement your ideas. Our framework is built based on pytorch_lightning, 
    and the structures is inspired by detectron2''',
    long_description_content_type="text/markdown",
    url="https://github.com/marsggbo/torchline",
    packages=find_packages(exclude=("tests", "projects")),
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)




#coding=utf-8
import torch
import numpy as np
import math
import random
import torchvision
from torchvision import transforms
from . import autoaugment
from torchline.utils import Registry, Logger

TRANSFORMS_REGISTRY = Registry('transforms')
TRANSFORMS_REGISTRY.__doc__ = """
Registry for data transform functions, i.e. torchvision.transforms

The registered object will be called with `obj(cfg)`
"""

LABEL_TRANSFORMS_REGISTRY = Registry('label_transforms')
LABEL_TRANSFORMS_REGISTRY.__doc__ = """
Registry for label transform functions, i.e. torchvision.transforms

The registered object will be called with `obj(cfg)`
"""

__all__ = [
    'build_transforms',
    'build_label_transforms',
    'TRANSFORMS_REGISTRY',
    'LABEL_TRANSFORMS_REGISTRY',
    'DefaultTransforms',
    'BaseTransforms'
]


def build_transforms(cfg):
    """
    Built the transforms, defined by `cfg.transforms.name`.
    """
    name = cfg.transforms.name
    return TRANSFORMS_REGISTRY.get(name)(cfg)
    
def build_label_transforms(cfg):
    """
    Built the label transforms, defined by `cfg.label_transforms.name`.
    """
    name = cfg.label_transforms.name
    if name == 'default':
        return None
    return LABEL_TRANSFORMS_REGISTRY.get(name)(cfg)

class BaseTransforms(object):
    def __init__(self, cfg):
        self.cfg = cfg
        self.logger_print = Logger(__name__, cfg).getlogger()
        self.is_train = cfg.dataset.is_train
        
    def get_transform(self):
        if not self.is_train: 
            self.logger_print.info('Generating validation transform ...')
            transform = self.valid_transform
            self.logger_print.info(f'Valid transform={transform}')
        else:
            self.logger_print.info('Generating training transform ...')
            transform = self.train_transform
            self.logger_print.info(f'Train transform={transform}')
        return transform

    @property
    def valid_transform(self):
        raise NotImplementedError

    @property
    def train_transform(self):
        raise NotImplementedError


@TRANSFORMS_REGISTRY.register()
class DefaultTransforms(BaseTransforms):
    def __init__(self, cfg):
        super(DefaultTransforms, self).__init__(cfg)
        self.is_train = cfg.dataset.is_train
        self.mean = cfg.transforms.tensor.normalization.mean
        self.std = cfg.transforms.tensor.normalization.std
        self.img_size = cfg.input.size
        self.padding = cfg.transforms.img.random_crop.padding
        self.min_edge_size = min(self.img_size)
        self.normalize = transforms.Normalize(self.mean, self.std)
        self.transform = self.get_transform()

    @property
    def valid_transform(self):
        transform = transforms.Compose([
            transforms.Resize(self.img_size),
            transforms.ToTensor(),
            self.normalize
        ])
        return transform

    @property
    def train_transform(self):
        # aug_imagenet
        if self.cfg.transforms.img.aug_imagenet:
            self.logger_print.info('Using imagenet augmentation')
            transform = transforms.Compose([
                transforms.Resize(self.img_size),
                autoaugment.ImageNetPolicy(),
                transforms.ToTensor(),
                self.normalize
            ])
        # aug cifar
        elif self.cfg.transforms.img.aug_cifar:
            self.logger_print.info('Using cifar augmentation')
            transform = transforms.Compose([
                transforms.Resize(self.img_size),
                autoaugment.CIFAR10Policy(),
                transforms.ToTensor(),
                self.normalize
            ])
        # customized transformations
        else:
            transform = self.read_transform_from_cfg()
        return transform

    def read_transform_from_cfg(self):
        transform_list = []
        self.check_conflict_options()
        img_transforms = self.cfg.transforms.img

        # resize and crop opertaion
        if img_transforms.random_resized_crop.enable:
            transform_list.append(transforms.RandomResizedCrop(self.img_size))
        elif img_transforms.resize.enable:
            transform_list.append(transforms.Resize(self.img_size))
        if img_transforms.random_crop.enable:
            transform_list.append(transforms.RandomCrop(self.min_edge_size, padding=self.padding))
        elif img_transforms.center_crop.enable:
            transform_list.append(transforms.CenterCrop(self.min_edge_size))

        # ColorJitter
        if img_transforms.color_jitter.enable:
            params = {key: img_transforms.color_jitter[key] for key in img_transforms.color_jitter 
                            if key != 'enable'}
            transform_list.append(transforms.ColorJitter(**params))

        # horizontal flip
        if img_transforms.random_horizontal_flip.enable:
            p = img_transforms.random_horizontal_flip.p
            transform_list.append(transforms.RandomHorizontalFlip(p))
        
        # vertical flip
        if img_transforms.random_vertical_flip.enable:
            p = img_transforms.random_vertical_flip.p
            transform_list.append(transforms.RandomVerticalFlip(p))

        # rotation
        if img_transforms.random_rotation.enable:
            degrees = img_transforms.random_rotation.degrees
            transform_list.append(transforms.RandomRotation(degrees))
        transform_list.append(transforms.ToTensor())
        transform_list.append(self.normalize)
        transform_list = transforms.Compose(transform_list)
        assert len(transform_list.transforms) > 0, "You must apply transformations"
        return transform_list

    def check_conflict_options(self):
        count = self.cfg.transforms.img.random_resized_crop.enable + \
                self.cfg.transforms.img.resize.enable
        assert count <= 1, 'You can only use one resize transform operation'

        count = self.cfg.transforms.img.random_crop.enable + \
                self.cfg.transforms.img.center_crop.enable
        assert count <= 1, 'You can only use one crop transform operation'